// genQuick.h -- Quick method prologue generation		-*- C++ -*-
// 
// Author: Ian.Piumarta@INRIA.Fr
//
// Last edited: Thu Jan 13 03:19:15 2000 by piumarta (Ian Piumarta) on pingu

#ifndef _j_genQuick_h
#define _j_genQuick_h


#include "archdep.h"

class NativeMethod;


extern void genPushSelf(NativeMethod *nMeth, insn *fail);
extern void genPushTrue(NativeMethod *nMeth, insn *fail);
extern void genPushFalse(NativeMethod *nMeth, insn *fail);
extern void genPushNil(NativeMethod *nMeth, insn *fail);
extern void genPushInteger(NativeMethod *nMeth, insn *fail);
extern void genLoadInstVar(NativeMethod *nMeth, insn *fail);


#endif // _j_genQuick_h

